//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#include <src/util.hpp>
#include <src/uri++.hpp>
#include <src/file_utils.hpp>
#include <src/vfs.hpp>
#include <src/database.hpp>
#include <src/library.hpp>
#include <src/main.hpp>
#include <src/x_library.hpp>

#include <glibmm.h>

namespace Bmp
{
  namespace VFS
  {
      class PluginContainerM3U
	  : public Bmp::VFS::PluginContainerBase
      {
	  public:

	    virtual bool
	    can_process (const std::string& uri)
	    {
		Bmp::URI u (uri);
		return str_has_suffix_nocase (u.path.c_str(), "m3u"); 
	    }

            virtual bool 
	    handle_read	(Bmp::VFS::Handle  &handle,
			 Util::FileList    &list)
	    {
		char **uris;
		int    n = 0;

		if (!handle.get_buffer()) throw Bmp::VFS::UNABLE_TO_PROCESS;

		uris = g_strsplit ((const char*)handle.get_buffer (), "\n", 0);
		 
		while (uris[n])
		{
			char *uri = uris[n];
			char *aux;

			//Strip leading spaces/tabs
			while ((uri[0] == ' ') || (uri[0] == '\t') || (uri[0] == '<')) uri++;

			//Skip comments, extended information or null lenght uris
			if ((uri[0] == '#') || (uri[0] == '\0'))
			    {
				n++;
				continue;
			    }
	      
			if (uri[0] == '/')
			  {
				  aux = g_strconcat ("file://", uri, NULL);
				  list.push_back (aux);
				  free (aux);
			  }
			else		
			  {
				  list.push_back (uri);
			  }
			n++;
		}

		g_strfreev (uris);
		return true;
	    }

	    virtual bool
	    can_write ()
	    {
	      return true;
	    }

	    virtual bool
	    handle_write  (Bmp::VFS::Handle&	  handle,
			   const Util::FileList&  uri_list)
	    {
		std::stringstream out;
		std::string outstr;
	
		out << "#EXTM3U\n";

		for (Util::FileList::const_iterator iter = uri_list . begin (); iter != uri_list . end (); ++iter)
		{
		      Bmp::DB::DataRow row;
		      try { library->get_metadata (*iter, row); } catch (Bmp::Library::Exception cxe) { continue; }
		      std::string artist, album, title;

		      Bmp::DB::DataRow::iterator i;
                      i = row.find (library->get_metadatum_id(Bmp::Library::DATUM_ARTIST));
                      if (i != row.end ())
                        artist = boost::get<Glib::ustring>(i->second);

                      i = row.find (library->get_metadatum_id(Bmp::Library::DATUM_ALBUM));
                      if (i != row.end ())
                        album = boost::get<Glib::ustring>(i->second);

                      i = row.find (library->get_metadatum_id(Bmp::Library::DATUM_TITLE));
                      if (i != row.end ())
                        title = boost::get<Glib::ustring>(i->second);

		      if ((artist.empty()) || (album.empty()) || (title.empty())) continue;

		      out << "#EXTINF:" << artist << "," << album << " - " << title << "\n";

		      Bmp::URI uri (*iter);

		      if (uri.get_protocol() != Bmp::URI::PROTOCOL_FILE)
			{ 		  
			  out << (*iter) << "\n";
			}
		      else	    
			{		      
			  uri.unescape ();
			  out << uri.path << "\n";
			}
		}
	       
		outstr = out.str(); 
		handle.set_buffer((const unsigned char *)strdup(outstr. c_str ()), strlen(outstr.c_str())+1);
		return true;
	    }

	    virtual Bmp::VFS::ExportData
	    get_export_data ()
	    {
	      static Bmp::VFS::ExportData export_data ("M3U Playlist", "m3u"); 
	      return export_data;
	    }
      };
  }
}
  
extern "C" Bmp::VFS::PluginContainerBase* plugin_create ()
{
  return new Bmp::VFS::PluginContainerM3U;
}

extern "C" void plugin_delete (Bmp::VFS::PluginContainerM3U* p)
{
  delete p;
}
