//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glibmm.h>
#include <glibmm/markup.h>

#include "uri++.hpp"
#include "debug.hpp"

#include "musicbrainz_xml.hpp"

#include "x_vfs.hpp"

namespace Bmp
{
  namespace MusicBrainzXML
  {
	    void
	    get_releases (const Glib::ustring& artist,
			  const Glib::ustring& album,
			  MbReleases& releases)
	    {
	      Glib::ustring uri;
	      uri.append ("http://musicbrainz.org/ws/1/release/?type=xml&title=").
		  append (album).
		  append ("&artist=").
		  append (artist).
		  append("&releasetypes=Official&limit=25");

	      Bmp::URI u (uri);
	      u.escape ();
	      Bmp::VFS::Handle handle = Bmp::VFS::Handle::create (Glib::ustring(u));

	      try
	      {
		  vfs->read (handle, Bmp::VFS::TRANSPORT);
		  MbReleaseParser parser;
		  Glib::Markup::ParseContext context (parser);

		  Glib::ustring data = reinterpret_cast<const char *>(handle.get_buffer()); 	  
		  context.parse (data);
		  context.end_parse ();
		  releases = parser.releases;
	      }
	      catch (Bmp::VFS::Exception& cxe)
	      {
		  //FIXME: VFS exception
	      }
	      catch (Glib::ConvertError& cxe)
	      {
		  //FIXME: Glib convert error (ustring) 
	      }
	      catch (Glib::MarkupError& cxe)
	      {
		  //FIXME: Glib Markup error
	      }
	    }
  
            void
            MbReleaseParser::on_start_element  (Glib::Markup::ParseContext& context,
                               const Glib::ustring& element_name,
                               const AttributeMap& attributes)
	    {
		if (element_name == "release")
		{
		  state |= E_RELEASE;
		  current = MbRelease();
		  current.release_id = (*(attributes.find ("id"))).second;
		  current.score = atoi ((*(attributes.find ("ext:score"))).second.c_str());
		  debug ("musicbrainz-xml", "<release id='%s' score='%d'>", current.release_id.c_str(), current.score);
		  return;
		}

		if (element_name == "title")
		{
		  state |= E_TITLE;
		  debug ("musicbrainz-xml", "<title>");
		  return;
		}

		if (element_name == "asin")
		{
		  state |= E_ASIN;
		  debug ("musicbrainz-xml", "<asin>");
		  return;
		}

		if (element_name == "artist")
		{
		  state |= E_ARTIST;
		  current.artist_id = (*(attributes.find ("id"))).second;
		  debug ("musicbrainz-xml", "<artist id='%s'>", current.artist_id.c_str());
		  return;
		}

		if (element_name == "name")
		{
		  state |= E_NAME;
		  debug ("musicbrainz-xml", "<name>");
		  return;
		}

		if (element_name == "disc-list")
		{
		  state |= E_DISC_LIST;
		  current.disc_list = atoi ((*(attributes.find ("count"))).second.c_str());
		  debug ("musicbrainz-xml", "<disc-list count='%d'>", current.disc_list);
		  return;
		}

		if (element_name == "track-list")
		{
		  state |= E_TRACK_LIST;
		  current.track_list = atoi ((*(attributes.find ("count"))).second.c_str());
		  debug ("musicbrainz-xml", "<track-list count='%d'>", current.track_list);
		  return;
		}

	    }

            void
            MbReleaseParser::on_end_element    (Glib::Markup::ParseContext& context,
                               const Glib::ustring& element_name)
	    {
		if (element_name == "release")
		{
		  state &= ~ E_RELEASE;
		  debug ("musicbrainz-xml", "</release>");

		  releases.push_back (current); 

		  if (state)
		  {
		    g_warning (G_STRLOC ": State should be 0, but is %d", state);
		    state = 0;
		  }
		  return;
		}

		if (element_name == "title")
		{
		  state &= ~ E_TITLE;
		  debug ("musicbrainz-xml", "</title>");
		  return;
		}

		if (element_name == "asin")
		{
		  state &= ~ E_ASIN;
		  debug ("musicbrainz-xml", "</asin>");
		  return;
		}

		if (element_name == "artist")
		{
		  state &= ~ E_ARTIST;
		  debug ("musicbrainz-xml", "</artist>");
		  return;
		}

		if (element_name == "name")
		{
		  state &= ~ E_NAME;
		  debug ("musicbrainz-xml", "</name>");
		  return;
		}

		if (element_name == "disc-list")
		{
		  state &= ~ E_DISC_LIST;
		  debug ("musicbrainz-xml", "</disc-list>");
		  return;
		}

		if (element_name == "track-list")
		{
		  state &= ~ E_TRACK_LIST;
		  debug ("musicbrainz-xml", "</track-list>");
		  return;
		}

	    }

            void
            MbReleaseParser::on_text           (Glib::Markup::ParseContext& context,
                               const Glib::ustring& text)
	    {
		if (state & E_TITLE)
		{
		  current.title = text;
		  debug ("musicbrainz-xml", "%s", text.c_str());
		  return;
		}

		if (state & E_ASIN)
		{
		  current.asin = text;
		  debug ("musicbrainz-xml", "%s", text.c_str());
		  return;
		}

		if ( (state & E_ARTIST) && (state & E_NAME) )
		{
		  current.artist_name = text;
		  debug ("musicbrainz-xml", "%s", text.c_str());
		  return;
		}
	    }

            void
            MbReleaseParser::on_passtrough     (Glib::Markup::ParseContext& context,
                               const Glib::ustring& text)
	    {
	    }

            void
            MbReleaseParser::on_error          (Glib::Markup::ParseContext& context,
                               const Glib::MarkupError& error)
	    {
	    }
    }
}

namespace Bmp
{
  namespace MusicBrainzXML
  {
	    void
	    get_tracks (const Glib::ustring& mb_albumid,
			MbTracks& tracks)
	    {
	      Glib::ustring uri;
	      uri.append ("http://musicbrainz.org/ws/1/release/").append(mb_albumid).append("?type=xml&inc=tracks");

	      Bmp::URI u (uri);
	      u.escape ();
	      Bmp::VFS::Handle handle = Bmp::VFS::Handle::create (Glib::ustring(u));
	      try
		{
		  vfs->read (handle, Bmp::VFS::TRANSPORT);
		  MbTrackParser parser;
		  Glib::Markup::ParseContext context (parser);

		  Glib::ustring data = reinterpret_cast<const char *>(handle.get_buffer()); 	  
		  context.parse (data);
		  context.end_parse ();
		  tracks = parser.tracks;
		}
	      catch (Bmp::VFS::Exception& cxe)
		{
		  //FIXME: VFS exception
		}
	      catch (Glib::ConvertError& cxe)
		{
		  //FIXME: Glib convert error (ustring) 
		}
	      catch (Glib::MarkupError& cxe)
		{
		  //FIXME: Glib Markup error
		}
	    }
  
            void
            MbTrackParser::on_start_element  (Glib::Markup::ParseContext& context,
                               const Glib::ustring& element_name,
                               const AttributeMap& attributes)
	    {
		if (element_name == "track")
		{
		  state |= E_TRACK;
		  current = MbTrack();
		  current.track_id = (*(attributes.find ("id"))).second;
		  current.tracknum = tracknum++;
		  debug ("musicbrainz-xml", "<track id='%s'>", current.track_id.c_str());
		  return;
		}

		if (element_name == "title")
		{
		  if (!(state & E_TRACK)) return;
		  state |= E_TITLE;
		  debug ("musicbrainz-xml", "<title>");
		  return;
		}

		if (element_name == "artist")
		{
		  if (!(state & E_TRACK)) return;
		  state |= E_ARTIST;
		  current.artist_id = (*(attributes.find ("id"))).second;
		  debug ("musicbrainz-xml", "<artist>");
		  return;
		}

		if (element_name == "name")
		{
		  if (!(state & E_TRACK)) return;
		  state |= E_NAME;
		  debug ("musicbrainz-xml", "<name>");
		  return;
		}
	    }

            void
            MbTrackParser::on_end_element    (Glib::Markup::ParseContext& context,
                               const Glib::ustring& element_name)
	    {
		if (element_name == "track")
		{
		  state &= ~ E_TRACK;
		  debug ("musicbrainz-xml", "</track>");
		  tracks.push_back (current); 

		  if (state)
		  {
		    g_warning (G_STRLOC ": State should be 0, but is %d", state);
		    state = 0;
		  }
		  return;
		}

		if (element_name == "title")
		{
		  state &= ~ E_TITLE;
		  debug ("musicbrainz-xml", "</title>");
		  return;
		}

		if (element_name == "artist")
		{
		  state &= ~ E_ARTIST;
		  debug ("musicbrainz-xml", "</artist>");
		  return;
		}

		if (element_name == "name")
		{
		  state &= ~ E_NAME;
		  debug ("musicbrainz-xml", "</name>");
		  return;
		}

	    }

            void
            MbTrackParser::on_text           (Glib::Markup::ParseContext& context,
                               const Glib::ustring& text)
	    {
		if (state & E_TITLE)
		{
		  current.title = text;
		  debug ("musicbrainz-xml", "%s", text.c_str());
		  return;
		}

		if (state & E_NAME)
		{
		  current.artist = text;
		  debug ("musicbrainz-xml", "%s", text.c_str());
		  return;
		}
	    }

            void
            MbTrackParser::on_passtrough     (Glib::Markup::ParseContext& context,
                               const Glib::ustring& text)
	    {
	    }

            void
            MbTrackParser::on_error          (Glib::Markup::ParseContext& context,
                               const Glib::MarkupError& error)
	    {
	    }
    }
}

